package gc;

import java.io.*;
import java.util.Vector;
import salsa.language.UniversalActor;
import salsa.language.Message;
/**
 * <p>Title: GC Object Input Stream</p>
 * <p>Description: This class resolves objects serialized by
 *                 ObjectOutputStream, and start GC
 *                 synchronization protocol if necessary.</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: wwc</p>
 * @author WeiJen
 * @version 1.0
 */

public class GCObjectInputStream extends ObjectInputStream {
  public static int SPLITTING_ONLY=0;
  public static int MUTE_GC=1;
  public static int ACTIVATE_GC=2;

  private int operationType=0;
  private WeakReference sourceRef;
  private Vector refSummary=null;
  private Vector mails=null;

  public GCObjectInputStream(InputStream in, int opType, WeakReference _ref) throws IOException,SecurityException{
    super(in);
    operationType=opType;
    sourceRef=_ref;
    refSummary=new Vector();
    mails=new Vector();
    enableResolveObject(true) ;
  }

  public GCObjectInputStream(InputStream in, boolean activateGC, WeakReference _ref) throws IOException,SecurityException{
    this( in, 0, _ref);
    if (activateGC) {operationType=ACTIVATE_GC;}
    else {operationType=MUTE_GC;}
  }

  public void setSourceReference(WeakReference _ref) {sourceRef=_ref;}

  public Vector getRefSummary() {return refSummary;}
  public void sendMails() {
    for (int i=0;i<mails.size();i++) {
      Message msg=(Message)mails.get(i);
      msg.getTarget().send(msg);
    }
    mails.clear();
  }


  /********
   * This function filters all actor references,
   * 1. case "in-transit to active": makes them GC active, and summarizes their information in
   *    refSummary. Lock the  target actor at this moment.
   * 2. case "active to in-transit": make them GC temporary silent, and summarizes their information in
   *    refSummary. Register the inverse acquaintance at this moment.
   * 3. Other case: it has the effect to create a copy only.
   ********/
  protected Object resolveObject(Object obj) {
    if (obj instanceof UniversalActor) {
      //case 1:
      //convert in-transit state to active GC state
      if (operationType==ACTIVATE_GC) {
         activateGCAndSendInvRefUnlockMsg((UniversalActor )obj);
      }

      //case 2:
      //convert active GC state to in-transit state
      else if (operationType==MUTE_GC) {
        muteGCAndSendLockMsg((UniversalActor )obj);
      }
    }
    return obj;
  }

  private void activateGCAndSendInvRefUnlockMsg(UniversalActor obj) {
    WeakReference ref=new WeakReference(obj);
    if ( obj.isInMessageState()) {
      obj.setSource(sourceRef);
      obj.toActivateGCState();
      if ( obj.getUAN() != null) {
        refSummary.addElement( obj.getUAN().toString());
      }
      else {
        refSummary.addElement( obj.getUAL().toString());
      }
      if (sourceRef!=null) {
        if (sourceRef.getUAN()!=null) {
          Object[] args={sourceRef.getUAN().toString()};
          //ref.send(new Message(sourceRef,ref,"putInvRefAndUnlock",args,null,null));
          mails.addElement(new Message(sourceRef,ref,"putInvRefAndUnlock",args,null,null));
        } else if (sourceRef.getUAL()!=null) {
          Object[] args={sourceRef.getUAL().toString()};
          //ref.send(new Message(sourceRef,ref,"putInvRefAndUnlock",args,null,null));
          mails.addElement(new Message(sourceRef,ref,"putInvRefAndUnlock",args,null,null));
        }
      }
    }
  }

  private void muteGCAndSendLockMsg(UniversalActor obj) {
    WeakReference ref=new WeakReference(obj);
    if ( obj.isInActiveGCState()) {
      obj.setSource(null);
      obj.toInMessageState();
      if (obj.getUAN() != null) {
        refSummary.addElement( obj.getUAN().toString());
      }
      else {
        refSummary.addElement(obj.getUAL().toString());
      }
      //lock the target actor
//System.out.println("$$$$$lock$$$$$: "+sourceRef + " -> " + ref);
      Object[] args={};
      ref.send(new Message(sourceRef,ref,"lock",args,null,null));
    }
  }
}